/*
 * NsisTask.java
 *
 * Copyright  2004 Wayne Grant
 * waynedgrant@hotmail.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * (This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package info.waynegrant.ant;

import java.io.*;
import java.text.MessageFormat;
import java.util.Vector;
import org.apache.tools.ant.*;
import org.apache.tools.ant.taskdefs.*;
import org.apache.tools.ant.types.*;

/**
 * Ant task to run the Nullsoft Scriptable Installation System (NSIS).
 * Calls the NSIS command-line exe which must be on the path.
 * Tested with NSIS version 2.0b3.
 *
 * @author Wayne Grant
 */
public class NsisTask extends Task
{
    /** NSIS command-line executable */
    private static final String NSIS_CMD_LINE = "makensis";

    /** NSIS command-line option for verbosity */
    private static final String NSIS_OPT_VERBOSITY = "/V";

    /** NSIS command-line option for output file */
    private static final String NSIS_OPT_OUTPUT = "/O";

    /** NSIS command-line option for pause after script execution */
    private static final String NSIS_OPT_PAUSE = "/PAUSE";

    /** NSIS command-line option for no inclusion of config header */
    private static final String NSIS_OPT_NOCONFIG = "/NOCONFIG";

    /** NSIS command-line option for no change directory */
    private static final String NSIS_OPT_NOCD = "/NOCD";

    /** NSIS command-line option for a define */
    private static final String NSIS_OPT_DEFINE = "/D";

    /** NSIS command-line option for a script command */
    private static final String NSIS_OPT_SCRIPTCMD = "/X";

    /** Default value for the verbosity attribute */
    private static final int DEFAULT_VERBOSITY = 0;

    /** Default value for the pause attribute */
    private static final boolean DEFAULT_PAUSE = false;

    /** Default value for the noconfig attribute */
    private static final boolean DEFAULT_NOCONFIG = false;

    /** Default value for the nocd attribute */
    private static final boolean DEFAULT_NOCD = false;

    /** Stores the value of the script attribute */
    private File m_fScript;

    /** Stores the value of the verbosity attribute */
    private int m_iVerbosity = DEFAULT_VERBOSITY;

    /** Stores the value of the out attribute */
    private File m_fOut;

    /** Stores the value of the pause attribute */
    private boolean m_bPause = DEFAULT_PAUSE;

    /** Stores the value of the noconfig attribute */
    private boolean m_bNoconfig = DEFAULT_NOCONFIG;

    /** Stores the value of the nocd attribute */
    private boolean m_bNocd = DEFAULT_NOCD;

    /** Stores nested define elements */
    private Vector m_vDefines = new Vector();

    /** Stores nested scriptcmd elements */
    private Vector m_vScriptcmds = new Vector();

    /**
     * Execute the task.
     *
     * @throws BuildException If a problem occurs while executing the task
     */
    public void execute() throws BuildException
    {
        // Check mandatory script attribute provided
        if (m_fScript == null)
        {
            throw new BuildException("script attribute required for nsis task", location);
        }

        // Set up arguments to NSIS command-line
        Vector vArgs = new Vector();

        vArgs.add(NSIS_OPT_VERBOSITY+m_iVerbosity); // Verbosity

        if (m_fOut != null)
        {
            vArgs.add(NSIS_OPT_OUTPUT+m_fOut.getPath()); // Out file?
        }

        if (m_bPause)
        {
            vArgs.add(NSIS_OPT_PAUSE); // Pause after script execution?
        }

        if (m_bNoconfig)
        {
            vArgs.add(NSIS_OPT_NOCONFIG); // No inclusion of nsisconf.nsh?
        }

        if (m_bNocd)
        {
            vArgs.add(NSIS_OPT_NOCD); // No change directory to script file directory?
        }

        // Define(s)?
        for (int iCnt=0; iCnt < m_vDefines.size(); iCnt++)
        {
            Define define = (Define)m_vDefines.get(iCnt);

            String sName = define.getName();
            String sValue = define.getValue();

            if (sName == null)
            {
                throw new BuildException("name attribute required for define element", location);
            }

            if (sValue == null)
            {
                throw new BuildException("value attribute required for define element", location);
            }

            vArgs.add(NSIS_OPT_DEFINE + sName + "=" + sValue);
        }

        // Script command(s)?
        for (int iCnt=0; iCnt < m_vScriptcmds.size(); iCnt++)
        {
            Scriptcmd scriptcmd = (Scriptcmd)m_vScriptcmds.get(iCnt);

            String sCmd = scriptcmd.getCmd();

            if (sCmd == null)
            {
                throw new BuildException("cmd attribute required for scriptcmd element", location);
            }

            vArgs.add(NSIS_OPT_SCRIPTCMD + sCmd);
        }

        vArgs.add(m_fScript.getPath()); // Script file

        // Create NSIS command-line
        Commandline cmd = new Commandline();
        cmd.setExecutable(NSIS_CMD_LINE);
        cmd.addArguments((String[])vArgs.toArray(new String[vArgs.size()]));

        // Create Execute object from command-line
        Execute exe = new Execute(new LogStreamHandler(this, Project.MSG_INFO, Project.MSG_WARN));
        exe.setAntRun(project);
        exe.setWorkingDirectory(project.getBaseDir());
        exe.setCommandline(cmd.getCommandline());

        try
        {
            // Execute command
            int iNsisRtn = exe.execute();

            if (iNsisRtn != 0)
            {
                throw new BuildException(MessageFormat.format("NSIS failed executing: ''{0}'', return code {1}", new String[]{cmd.toString(), ""+iNsisRtn}), location);
            }
        }
        catch (IOException ex)
        {
            ex.printStackTrace();
            throw new BuildException(MessageFormat.format("NSIS failed executing: ''{0}''", new String[]{cmd.toString()}), ex, location);
        }
    }

    /**
     * Set the script attribute.
     *
     * @param fScript Attribute value
     */
    public void setScript(File fScript)
    {
        m_fScript = fScript;
    }

    /**
     * Set the verbosity attribute.
     *
     * @param verbosity Attribute value
     */
    public void setVerbosity(Verbosity verbosity)
    {
        try
        {
            m_iVerbosity = Integer.parseInt(verbosity.getValue());
        }
        catch (NumberFormatException ex)
        {
            // Cannot happen
            ex.printStackTrace();
        }
    }

    /**
     * Enumerated attribute for verbosity: 0-4.
     */
    public static class Verbosity extends EnumeratedAttribute
    {
        /**
         * Get enumeration values.
         *
         * @return Enumeration values
         */
        public String[] getValues()
        {
            return new String[] {"0", "1", "2", "3", "4"};
        }
    }

    /**
     * Set the out attribute.
     *
     * @param fOut Attribute value
     */
    public void setOut(File fOut)
    {
        m_fOut = fOut;
    }

    /**
     * Set the pause attribute.
     *
     * @param bPause Attribute value
     */
    public void setPause(boolean bPause)
    {
        m_bPause = bPause;
    }

    /**
     * Set the noconfig attribute.
     *
     * @param bNoconfig Attribute value
     */
    public void setNoconfig(boolean bNoconfig)
    {
        m_bNoconfig = bNoconfig;
    }

    /**
     * Set the nocd attribute.
     *
     * @param bNocd Attribute value
     */
    public void setNocd(boolean bNocd)
    {
        m_bNocd = bNocd;
    }

    /**
     * Add define nested element.
     *
     * @param define Nested element
     */
    public void addConfiguredDefine(Define define)
    {
        m_vDefines.add(define);
    }

    /**
     * Nested element define.
     */
    public static class Define extends Object
    {
        /** Stores the value of the name attribute */
        private String m_sName;

        /** Stores the value of the value attribute */
        private String m_sValue;

        /**
         * Set the name attribute.
         *
         * @param sName Attribute value
         */
        public void setName(String sName)
        {
            m_sName = sName;
        }

        /**
         * Get the name attribute.
         *
         * @return Attribute value
         */
        String getName() // Deliberately package private
        {
            return m_sName;
        }

        /**
         * Set the value attribute.
         *
         * @param sValue Attribute value
         */
        public void setValue(String sValue)
        {
            m_sValue= sValue;
        }

        /**
         * Get the value attribute.
         *
         * @return Attribute value
         */
        String getValue()
        {
            return m_sValue;
        }
    }

    /**
     * Add scriptcmd nested element.
     *
     * @param scriptcmd Nested element
     */
    public void addConfiguredScriptcmd(Scriptcmd scriptcmd)
    {
        m_vScriptcmds.add(scriptcmd);
    }

    /**
     * Nested element scriptcmd.
     */
    public static class Scriptcmd extends Object
    {
        /** Stores the value of the cmd attribute */
        private String m_sCmd;

        /**
         * Set the cmd attribute.
         *
         * @param sCmd Attribute value
         */
        public void setCmd(String sCmd)
        {
            m_sCmd = sCmd;
        }

        /**
         * Get the cmd attribute.
         *
         * @return Attribute value
         */
        String getCmd() // Deliberately package private
        {
            return m_sCmd;
        }
    }
}
